! Example 3 of the MPI course. It performs parallelized 
! calculation of the multiplication of  
!     matrix A with matrix B.
! HPCVL, Queen's University, 2006

MODULE MPI
  INCLUDE 'mpif.h'
  INTEGER::MYID,TOTPS, IERR
END MODULE MPI

MODULE MATRIXES
  INTEGER:: MSZ(3),IB,IE,AL,RL
  REAL*8, ALLOCATABLE :: &
          MP(:,:),MA(:,:),MB(:,:),MC(:,:)
END MODULE MATRIXES

PROGRAM EXAMPLE03
  USE   MPI
  CALL  MPIINIT
  CALL  MTXINIT
  CALL  DEMO03
  WRITE(*,*)' PROCESS ',MYID,' OF ',TOTPS,' DONE '
  CALL  MPI_FINALIZE(IERR)
  STOP
END PROGRAM EXAMPLE03

SUBROUTINE MPIINIT
  USE  MPI
  CALL MPI_INIT( IERR )
  CALL MPI_COMM_RANK(MPI_COMM_WORLD,MYID,IERR)
  CALL MPI_COMM_SIZE(MPI_COMM_WORLD,TOTPS,IERR)
  RETURN
END SUBROUTINE MPIINIT

SUBROUTINE MTXINIT
  USE MPI
  USE MATRIXES

  IF(MYID.EQ.0) THEN
     OPEN(1,FILE='in.dat')
     READ(1,*) MSZ(3), MSZ(2), MSZ(1)
  END IF

  CALL MPI_BCAST(MSZ,3,MPI_INTEGER,0,MPI_COMM_WORLD,IERR)

  AL=(MSZ(3)-1)/TOTPS+1
  IB=1+MYID*AL
  IE=MIN(IB+AL-1,MSZ(3))

  ALLOCATE(MA(MSZ(2),MSZ(3)))
  ALLOCATE(MB(MSZ(2),MSZ(1)))
  ALLOCATE(MP(MSZ(3),MSZ(1)))
  ALLOCATE(MC(MSZ(3),MSZ(1)))

  IF(MYID.EQ.0) THEN
     READ(1,*)((MA(J,I),J=1,MSZ(2)),I=1,MSZ(3))
     READ(1,*)((MB(I,J),J=1,MSZ(1)),I=1,MSZ(2))
     CLOSE(1)
  END IF

  CALL MPI_BCAST(MA,MSZ(2)*MSZ(3),MPI_REAL8,0,MPI_COMM_WORLD,IERR)
  CALL MPI_BCAST(MB,MSZ(2)*MSZ(1),MPI_REAL8,0,MPI_COMM_WORLD,IERR)

  RETURN
END SUBROUTINE MTXINIT




SUBROUTINE DEMO03
  USE MPI
  USE MATRIXES

  MP=0.0D0
  DO I=IB,IE
  DO J=1,MSZ(1)
  DO K=1,MSZ(2)
     MP(I,J)=MP(I,J)+MA(K,I)*MB(K,J)
  END DO
  END DO
  END DO

  MC=0.0D0
  CALL MPI_REDUCE(MP,MC, MSZ(3)*MSZ(1), &
   MPI_REAL8,MPI_SUM,0,MPI_COMM_WORLD,IERR)

  IF(MYID.EQ.0) THEN
     OPEN(1,FILE='out.dat')
     DO J=1, MSZ(3)
        WRITE(1,*)(MC(J,I),I=1,MSZ(1))
     END DO
     CLOSE(1)
  END IF

  DEALLOCATE(MA,MB,MP,MC)

  RETURN
END SUBROUTINE DEMO03




