module mpi
  include 'mpif.h'
  integer::myid,totps, ierr
end module mpi

module matrixes
  integer:: msz(3),ib,ie,al,rl
  real*8, allocatable :: mp(:,:),ma(:,:),mb(:,:),mc(:,:)
  real*8 :: oosm
end module matrixes

program example03
  use   mpi
! Initialization
  if (myid.eq.0) call cpu_time (t1)
  call  mpiinit
  call  mtxinit
  if (myid.eq.0) then
     call cpu_time (t2)
     write (*,*)' Time for initialization:',t2-t1,' seconds'
  end if
! Computation
  call  demo03
  if (myid.eq.0) then
     call cpu_time(t1)
     write (*,*) ' Time for computation:',t1-t2,' seconds'
  end if
 ! Finalization 
  call  mpi_finalize(ierr)
  stop
end program example03

subroutine mpiinit
  use  mpi
  call mpi_init( ierr )
  call mpi_comm_rank(mpi_comm_world,myid,ierr)
  call mpi_comm_size(mpi_comm_world,totps,ierr)
  return
end subroutine mpiinit

subroutine mtxinit
  use mpi
  use matrixes

  if(myid.eq.0) then
     open(1,file='dim.dat')
     read(1,*) msz(3), msz(2), msz(1)
     close(1)
     oosm=1/sqrt(dble(msz(2)))
  end if

  call mpi_bcast(msz, 3, mpi_integer, 0, mpi_comm_world,ierr)

  al=(msz(3)-1)/totps+1
  ib=1+myid*al
  ie=min(ib+al-1,msz(3))

  allocate(ma(msz(2),msz(3)))
  allocate(mb(msz(2),msz(1)))
  allocate(mp(msz(3),msz(1)))
  allocate(mc(msz(3),msz(1)))

  if (myid.eq.0) call setmat(1)
  if (myid.eq.0) call setmat(2)

  call mpi_bcast(ma, msz(2)*msz(3), mpi_real8, 0, mpi_comm_world,ierr)
  call mpi_bcast(mb, msz(2)*msz(1), mpi_real8, 0, mpi_comm_world,ierr)

  return
end subroutine mtxinit

subroutine demo03
  use mpi
  use matrixes

  mp=0.0d0
  do i=ib,ie
     do j=1,msz(1)
        do k=1,msz(2)
           mp(i,j)=mp(i,j)+ma(k,i)*mb(k,j)
        end do
     end do
  end do
  
  mc=0.0d0
  call mpi_reduce(mp,mc, msz(3)*msz(1), mpi_real8, mpi_sum, 0, mpi_comm_world, ierr)
!  if (myid.eq.0) write(*,*) mc
  deallocate(ma,mb,mp,mc)

  return
end subroutine demo03

subroutine setmat(iset)
  use matrixes
  if (iset.eq.1) then 
     ma=oosm
  else if (iset.eq.2) then 
     mb=oosm
  else
     write(*,*) ' Invalid value for iset = ',iset
     stop
  end if
end subroutine setmat

