! Example 4 of the MPI course. It does the same 
! calculation as that of the example 3, but it 
! will reduce much memory. Or the memory of 
! some matrixes are distributed. Also in this  
! example, both blocking and non-blocking 
! point-to-point communications are demonstrated.
! HPCVL, Queen's University, 2006


MODULE MPI
  INCLUDE 'mpif.h'
  INTEGER::MYID,TOTPS, IERR
END MODULE MPI


MODULE MATRIXES
  INTEGER              :: MSZ(3),AL
  INTEGER, ALLOCATABLE :: IBE(:,:)
  REAL*8,  ALLOCATABLE :: &
             MP(:,:),MA(:,:),MB(:,:),MC(:,:)
END MODULE MATRIXES




PROGRAM EXAMPLE04
  USE   MPI
  CALL  MPIINIT
  CALL  MTXINIT
  CALL  DEMO04
  WRITE(*,*)' PROCESS ',MYID,' OF ',TOTPS,' DONE '
  CALL  MPI_FINALIZE(IERR)
  STOP
END PROGRAM EXAMPLE04




SUBROUTINE MPIINIT
  USE  MPI
  CALL MPI_INIT( IERR )
  CALL MPI_COMM_RANK(MPI_COMM_WORLD,MYID,IERR)
  CALL MPI_COMM_SIZE(MPI_COMM_WORLD,TOTPS,IERR)
  RETURN
END SUBROUTINE MPIINIT




SUBROUTINE DSTRBT(NP,NW,AL,IBE)
  INTEGER::AL,I,J,IC,NP,NW,IBE(3,0:NP-1)
  AL=(NW-1)/NP+1
  DO IC=0,NP-1
     IBE(1,IC)=IC*AL+1
     IBE(2,IC)=MIN(IC*AL+AL,NW)
     IBE(3,IC)=MAX(0,IBE(2,IC)-IBE(1,IC)+1)
  END DO
  RETURN
END SUBROUTINE DSTRBT




SUBROUTINE MTXINIT
  USE MPI
  USE MATRIXES
  INTEGER:: MPISTTS(MPI_STATUS_SIZE)
  INTEGER:: MPITAG, I, J, IC

  IF(MYID.EQ.0) THEN
     OPEN(1,FILE='in.dat')
     READ(1,*) MSZ(3), MSZ(2), MSZ(1)
  END IF

  CALL MPI_BCAST(MSZ,3,MPI_INTEGER,0, &
                 MPI_COMM_WORLD,IERR)

  ALLOCATE(IBE(3,0:TOTPS-1))

  CALL DSTRBT(TOTPS,MSZ(3),AL,IBE)

  ALLOCATE(MP(MSZ(1),MAX(1,IBE(3,MYID))))
  ALLOCATE(MA(MSZ(2),MAX(1,IBE(3,MYID))))
  ALLOCATE(MB(MSZ(2),MSZ(1)))

  MPITAG=32700

  IF(MYID.EQ.0) THEN
     READ(1,*)((MA(J,I),J=1,MSZ(2)),I=1,IBE(3,0))
           
     ALLOCATE(MC(MSZ(2),AL))
     DO IC=1,TOTPS-1
        DO I=1,IBE(3,IC)
           READ(1,*)(MC(J,I),J=1,MSZ(2))
        END DO
        CALL MPI_SEND(MC,MSZ(2)*IBE(3,IC),   &
                         MPI_REAL8,IC,MPITAG,&
                         MPI_COMM_WORLD, IERR)
     END DO
     DEALLOCATE(MC)
  ELSE
     CALL MPI_RECV(MA,MSZ(2)*IBE(3,MYID),    &
                  MPI_REAL8,0, MPITAG,       &
                  MPI_COMM_WORLD,MPISTTS,IERR)
  END IF

  IF(MYID.EQ.0) THEN
     READ(1,*)((MB(I,J),J=1,MSZ(1)),I=1,MSZ(2))
     CLOSE(1)
  END IF
  CALL MPI_BCAST(MB,MSZ(2)*MSZ(1), MPI_REAL8, &
                        0, MPI_COMM_WORLD,IERR)

  RETURN
END SUBROUTINE MTXINIT




SUBROUTINE DEMO04
  USE MPI
  USE MATRIXES
  INTEGER:: MPIRQSTS,MPIRQSTR(TOTPS)
  INTEGER:: MPISTTS(MPI_STATUS_SIZE)
  INTEGER:: MPITAG, I, J, K, IC

  MP=0.0D0
  DO I=1,IBE(3,MYID)
  DO J=1,MSZ(1)
  DO K=1,MSZ(2)
     MP(J,I)=MP(J,I)+MA(K,I)*MB(K,J)
  END DO
  END DO
  END DO

  IF(MYID.EQ.0) THEN
     ALLOCATE(MC(MSZ(1),MSZ(3)))
     MC=0.0D0
     MC(1:MSZ(1),1:IBE(3,0))=MP(1:MSZ(1),1:IBE(3,0))
  END IF

  MPITAG=100+MYID

  IF(MYID.EQ.0) THEN
     DO IC=1,TOTPS-1
        CALL MPI_IRECV(MC(1, IBE(1,IC)),     &
                 MSZ(1)*IBE(3,IC),MPI_REAL8, & 
                 IC,100+IC,MPI_COMM_WORLD,   &
                 MPIRQSTR(IC),IERR)
     END DO
  ELSE
     CALL MPI_ISEND(MP, MSZ(1)*IBE(3,MYID),  &
                 MPI_REAL8,0, MPITAG,        &
                 MPI_COMM_WORLD,MPIRQSTS,IERR)
  END IF

  IF(MYID.EQ.0) THEN
     DO IC=1,TOTPS-1
        CALL MPI_WAIT(MPIRQSTR(IC),MPISTTS,IERR)
     END DO
  ELSE
     CALL MPI_WAIT(MPIRQSTS,MPISTTS,IERR)
  END IF

  IF(MYID.EQ.0) THEN
     OPEN(1,FILE='out.dat')
         DO J=1, MSZ(3)
            WRITE(1,*)(MC(I,J),I=1,MSZ(1))
         END DO
     CLOSE(1)
  END IF

  DEALLOCATE(MA,MB,MP,IBE)
  IF(MYID.EQ.0) THEN
     DEALLOCATE(MC)
  END IF

  RETURN
END SUBROUTINE DEMO04


