module mpi
  include 'mpif.h'
  integer::myid,totps, ierr
end module mpi

module matrixes
  integer              :: msz(3),al
  integer, allocatable :: ibe(:,:)
  real*8,  allocatable :: mp(:,:),ma(:,:),mb(:,:),mc(:,:)
  real*8 :: oosm
end module matrixes

program example04
  use   mpi
! Initialization
  if (myid.eq.0) call cpu_time (t1)
  call  mpiinit
  call  mtxinit
  if (myid.eq.0) then
     call cpu_time (t2)
     write (*,*)' Time for initialization:',t2-t1,' seconds'
  end if
! Computation
  call  demo04
  if (myid.eq.0) then
     call cpu_time(t1)
     write (*,*) ' Time for computation:',t1-t2,' seconds'
  end if
 ! Finalization 
  call  mpi_finalize(ierr)
  stop
end program example04

subroutine mpiinit
  use  mpi
  call mpi_init( ierr )
  call mpi_comm_rank(mpi_comm_world,myid,ierr)
  call mpi_comm_size(mpi_comm_world,totps,ierr)
  return
end subroutine mpiinit

subroutine dstrbt(np,nw,al,ibe)
  integer::al,i,j,ic,np,nw,ibe(3,0:np-1)
  al=(nw-1)/np+1
  do ic=0,np-1
     ibe(1,ic)=ic*al+1
     ibe(2,ic)=min(ic*al+al,nw)
     ibe(3,ic)=max(0,ibe(2,ic)-ibe(1,ic)+1)
  end do
  return
end subroutine dstrbt

subroutine mtxinit
  use mpi
  use matrixes
  integer:: mpistts(mpi_status_size)
  integer:: mpitag, i, j, ic

  if(myid.eq.0) then
     open(1,file='dim.dat')
     read(1,*) msz(3), msz(2), msz(1)
     close(1)
     oosm=1/sqrt(dble(msz(2)))
  end if

  call mpi_bcast(msz,3,mpi_integer,0, mpi_comm_world,ierr)

  allocate(ibe(3,0:totps-1))

  call dstrbt(totps,msz(3),al,ibe)

  allocate(mp(msz(1),max(1,ibe(3,myid))))
  allocate(ma(msz(2),max(1,ibe(3,myid))))
  allocate(mb(msz(2),msz(1)))

  mpitag=32700

  if(myid.eq.0) then
     call setmat(1)
     allocate(mc(msz(2),al))
     do ic=1,totps-1
        call setmat(3)
        call mpi_send(mc,msz(2)*ibe(3,ic), mpi_real8,ic,mpitag,mpi_comm_world, ierr)
     end do
     deallocate(mc)
  else
     call mpi_recv(ma,msz(2)*ibe(3,myid), mpi_real8,0, mpitag, mpi_comm_world,mpistts,ierr)
  end if

  if(myid.eq.0) call setmat(2)
  call mpi_bcast(mb,msz(2)*msz(1), mpi_real8, 0, mpi_comm_world,ierr)

  return
end subroutine mtxinit

subroutine demo04
  use mpi
  use matrixes
  integer:: mpirqsts,mpirqstr(totps)
  integer:: mpistts(mpi_status_size)
  integer:: mpitag, i, j, k, ic

  mp=0.0d0
! call cpu_time(t1)
  do i=1,ibe(3,myid)
     do j=1,msz(1)
        do k=1,msz(2)
           mp(j,i)=mp(j,i)+ma(k,i)*mb(k,j)
        end do
     end do
  end do
! call cpu_time(t2)
! write(*,*)' Pure computation time on process', myid, ' = ',t2-t1,' seconds'

  if(myid.eq.0) then
     allocate(mc(msz(1),msz(3)))
     mc=0.0d0
     mc(1:msz(1),1:ibe(3,0))=mp(1:msz(1),1:ibe(3,0))
  end if

  mpitag=100+myid

  if(myid.eq.0) then
     do ic=1,totps-1
        call mpi_irecv(mc(1, ibe(1,ic)), msz(1)*ibe(3,ic),mpi_real8, & 
                 ic,100+ic,mpi_comm_world, mpirqstr(ic),ierr)
     end do
  else
     call mpi_isend(mp, msz(1)*ibe(3,myid), mpi_real8,0, mpitag, &
                 mpi_comm_world,mpirqsts,ierr)
  end if

  if(myid.eq.0) then
     do ic=1,totps-1
        call mpi_wait(mpirqstr(ic),mpistts,ierr)
     end do
  else
     call mpi_wait(mpirqsts,mpistts,ierr)
  end if

! if(myid.eq.0) write(*,*) mc

  deallocate(ma,mb,mp,ibe)
  if(myid.eq.0) deallocate(mc)

  return
end subroutine demo04

subroutine setmat(iset)
  use matrixes
  if (iset.eq.1) then 
     ma=oosm
  else if (iset.eq.2) then
     mb=oosm
  else if (iset.eq.3) then
     mc=oosm
  else
     write(*,*) ' Invalid value for iset = ',iset
     stop
  end if
end subroutine setmat

