! Example 5 of the MPI course. It does the same 
! calculation as that of the example 3, but it 
! will reduce much memory. Or the memory of 
! some matrixes are distributed. Also in this  
! example, we use MPI_SCATTERV and MPI_GATHERV
! collective communications to replace the
! point-to-point communications in example 4.
! HPCVL, Queen's University, 2006


MODULE MPI
  INCLUDE 'mpif.h'
  INTEGER::MYID,TOTPS, IERR
END MODULE MPI


MODULE MATRIXES
  INTEGER              :: MSZ(3),AL
  INTEGER, ALLOCATABLE :: IBE(:,:)
  REAL*8,  ALLOCATABLE :: &
             MP(:,:),MA(:,:),MB(:,:),MC(:,:)
END MODULE MATRIXES




PROGRAM EXAMPLE05
  USE   MPI
  CALL  MPIINIT
  CALL  MTXINIT
  CALL  DEMO05
  WRITE(*,*)' PROCESS ',MYID,' OF ',TOTPS,' DONE '
  CALL  MPI_FINALIZE(IERR)
  STOP
END PROGRAM EXAMPLE05




SUBROUTINE MPIINIT
  USE  MPI
  CALL MPI_INIT( IERR )
  CALL MPI_COMM_RANK(MPI_COMM_WORLD,MYID,IERR)
  CALL MPI_COMM_SIZE(MPI_COMM_WORLD,TOTPS,IERR)
  RETURN
END SUBROUTINE MPIINIT




SUBROUTINE DSTRBT(NP,NW,AL,IBE)
  INTEGER::AL,I,J,IC,NP,NW,IBE(3,0:NP-1)
  AL=(NW-1)/NP+1
  DO IC=0,NP-1
     IBE(1,IC)=IC*AL+1
     IBE(2,IC)=MIN(IC*AL+AL,NW)
     IBE(3,IC)=MAX(0,IBE(2,IC)-IBE(1,IC)+1)
  END DO
  RETURN
END SUBROUTINE DSTRBT




SUBROUTINE MTXINIT
  USE MPI
  USE MATRIXES
  INTEGER:: I, J, IC
  INTEGER:: MPIRECVC(0:TOTPS-1)
  INTEGER:: MPIDISPLS(0:TOTPS-1)
  INTEGER:: MPISTTS(MPI_STATUS_SIZE)

  IF(MYID.EQ.0) THEN
     OPEN(1,FILE='in.dat')
     READ(1,*) MSZ(3), MSZ(2), MSZ(1)
  END IF

  CALL MPI_BCAST(MSZ,3,MPI_INTEGER,0, & 
                   MPI_COMM_WORLD,IERR)

  ALLOCATE(IBE(3,0:TOTPS-1))

  CALL DSTRBT(TOTPS,MSZ(3),AL,IBE)

  ALLOCATE(MP(MSZ(1),MAX(1,IBE(3,MYID))))
  ALLOCATE(MA(MSZ(2),MAX(1,IBE(3,MYID))))
  ALLOCATE(MB(MSZ(2),MSZ(1)))

  IF(MYID.EQ.0) THEN
     ALLOCATE(MC(MSZ(2),MSZ(3)))
  END IF

  IF(MYID.EQ.0) THEN
     READ(1,*)((MC(J,I),J=1,MSZ(2)),I=1,MSZ(3))
  END IF

  DO I=0, TOTPS-1
      MPIRECVC(I)=IBE(3,I)*MSZ(2)
      MPIDISPLS(I)=(IBE(1,I)-1)*MSZ(2)
  END DO

  CALL MPI_SCATTERV(MC,MPIRECVC, MPIDISPLS, &
           MPI_REAL8,MA,MSZ(2)*IBE(3,MYID), &
           MPI_REAL8,0, MPI_COMM_WORLD,IERR)

  IF(MYID.EQ.0) THEN
     DEALLOCATE(MC)
  END IF

  IF(MYID.EQ.0) THEN
     READ(1,*)((MB(I,J),J=1,MSZ(1)),I=1,MSZ(2))
     CLOSE(1)
  END IF
  CALL MPI_BCAST(MB,MSZ(2)*MSZ(1),MPI_REAL8,0,&
                           MPI_COMM_WORLD,IERR)

  RETURN
END SUBROUTINE MTXINIT




SUBROUTINE DEMO05
  USE MPI
  USE MATRIXES
  INTEGER:: MPIRECVC(0:TOTPS-1)
  INTEGER:: MPIDISPLS(0:TOTPS-1)
  INTEGER:: I, J, K, IC

  MP=0.0D0
  DO I=1,IBE(3,MYID)
  DO J=1,MSZ(1)
  DO K=1,MSZ(2)
     MP(J,I)=MP(J,I)+MA(K,I)*MB(K,J)
  END DO
  END DO
  END DO

  IF(MYID.EQ.0) THEN
     ALLOCATE(MC(MSZ(1),MSZ(3)))
  END IF

  DO I=0, TOTPS-1
     MPIRECVC(I)=IBE(3,I)*MSZ(1)
     MPIDISPLS(I)=(IBE(1,I)-1)*MSZ(1)
  END DO

  CALL MPI_GATHERV(MP, MSZ(1)*IBE(3,MYID), &
          MPI_REAL8,MC,MPIRECVC,MPIDISPLS, &
          MPI_REAL8,0, MPI_COMM_WORLD,IERR)

  IF(MYID.EQ.0) THEN
     OPEN(1,FILE='out.dat')
        DO J=1, MSZ(3)
           WRITE(1,*)(MC(I,J),I=1,MSZ(1))
        END DO
     CLOSE(1)
  END IF

  DEALLOCATE(MA,MB,MP,IBE)
  IF(MYID.EQ.0) THEN
      DEALLOCATE(MC)
  END IF

  RETURN
END SUBROUTINE DEMO05



