module mpi
  include 'mpif.h'
  integer::myid,totps, ierr
end module mpi

module matrixes
  integer              :: msz(3),al
  integer, allocatable :: ibe(:,:)
  real*8,  allocatable :: mp(:,:),ma(:,:),mb(:,:),mc(:,:)
  real*8 :: oosm
end module matrixes

program example05
  use   mpi
! Initialization
  if (myid.eq.0) call cpu_time (t1)
  call  mpiinit
  call  mtxinit
  if (myid.eq.0) then
     call cpu_time (t2)
     write (*,*)' Time for initialization:',t2-t1,' seconds'
  end if
! Computation
  call  demo05
  if (myid.eq.0) then
     call cpu_time(t1)
     write (*,*) ' Time for computation:',t1-t2,' seconds'
  end if
 ! Finalization 
  call  mpi_finalize(ierr)
  stop
end program example05

subroutine mpiinit
  use  mpi
  call mpi_init( ierr )
  call mpi_comm_rank(mpi_comm_world,myid,ierr)
  call mpi_comm_size(mpi_comm_world,totps,ierr)
  return
end subroutine mpiinit

subroutine dstrbt(np,nw,al,ibe)
  integer::al,i,j,ic,np,nw,ibe(3,0:np-1)
  al=(nw-1)/np+1
  do ic=0,np-1
     ibe(1,ic)=ic*al+1
     ibe(2,ic)=min(ic*al+al,nw)
     ibe(3,ic)=max(0,ibe(2,ic)-ibe(1,ic)+1)
  end do
  return
end subroutine dstrbt

subroutine mtxinit
  use mpi
  use matrixes
  integer:: i, j, ic
  integer:: mpirecvc(0:totps-1)
  integer:: mpidispls(0:totps-1)
  integer:: mpistts(mpi_status_size)

  if(myid.eq.0) then
     open(1,file='dim.dat')
     read(1,*) msz(3), msz(2), msz(1)
     close(1)
     oosm=1/sqrt(dble(msz(2)))
  end if

  call mpi_bcast(msz,3,mpi_integer,0, mpi_comm_world,ierr)

  allocate(ibe(3,0:totps-1))

  call dstrbt(totps,msz(3),al,ibe)

  allocate(mp(msz(1),max(1,ibe(3,myid))))
  allocate(ma(msz(2),max(1,ibe(3,myid))))
  allocate(mb(msz(2),msz(1)))

  if(myid.eq.0) allocate(mc(msz(2),msz(3)))

  if(myid.eq.0) call setmat(3)

  do i=0, totps-1
      mpirecvc(i)=ibe(3,i)*msz(2)
      mpidispls(i)=(ibe(1,i)-1)*msz(2)
  end do

  call mpi_scatterv(mc,mpirecvc, mpidispls, &
           mpi_real8,ma,msz(2)*ibe(3,myid), &
           mpi_real8,0, mpi_comm_world,ierr)

  if(myid.eq.0) then
     deallocate(mc)
  end if

  if(myid.eq.0) call setmat(2)

  call mpi_bcast(mb,msz(2)*msz(1),mpi_real8,0,mpi_comm_world,ierr)

  return
end subroutine mtxinit

subroutine demo05
  use mpi
  use matrixes
  integer:: mpirecvc(0:totps-1)
  integer:: mpidispls(0:totps-1)
  integer:: i, j, k, ic

  mp=0.0d0
  do i=1,ibe(3,myid)
  do j=1,msz(1)
  do k=1,msz(2)
     mp(j,i)=mp(j,i)+ma(k,i)*mb(k,j)
  end do
  end do
  end do

  if(myid.eq.0) then
     allocate(mc(msz(1),msz(3)))
  end if

  do i=0, totps-1
     mpirecvc(i)=ibe(3,i)*msz(1)
     mpidispls(i)=(ibe(1,i)-1)*msz(1)
  end do

  call mpi_gatherv(mp, msz(1)*ibe(3,myid), &
          mpi_real8,mc,mpirecvc,mpidispls, &
          mpi_real8,0, mpi_comm_world,ierr)

  ! if(myid.eq.0) write(*,*) mc

  deallocate(ma,mb,mp,ibe)
  if(myid.eq.0) deallocate(mc)

  return
end subroutine demo05

subroutine setmat(iset)
  use matrixes
  if (iset.eq.1) then 
     ma=oosm
  else if (iset.eq.2) then
     mb=oosm
  else if (iset.eq.3) then
     mc=oosm
  else
     write(*,*) ' Invalid value for iset = ',iset
     stop
  end if
end subroutine setmat



