! This is an example code for the Master-Slave 
! Parallel Model.
! HPCVL, Queen's University, 2006

module work
  integer, dimension(20) :: sleeptime
contains
  subroutine load
    integer :: i
    real :: r
    do i=1,20
       call random_number(r)
       sleeptime(i)=1+int(4*r)
    end do
  end subroutine load
end module work

MODULE MPI
  INCLUDE 'mpif.h'
  INTEGER::MYID,TOTPS, IERR
END MODULE MPI

MODULE JOBDSTRB
  INTEGER:: NJOB
  INTEGER, ALLOCATABLE:: IDID(:)
  INTEGER, ALLOCATABLE:: SIDID(:)
  INTEGER, ALLOCATABLE:: NIDID(:)
  INTEGER, ALLOCATABLE:: SNIDID(:)
  INTEGER, ALLOCATABLE:: IFNSH(:)
  INTEGER, ALLOCATABLE:: SIFNSH(:)
  INTEGER, ALLOCATABLE:: IGAVE(:)
  INTEGER, ALLOCATABLE:: PFNSH(:)
  INTEGER, ALLOCATABLE:: NIGAVE(:)
END MODULE JOBDSTRB

PROGRAM EXAMPLE21
  USE   MPI
  CALL  MPIINIT
  CALL  JOBINIT
  CALL  DEMO21
  CALL  MPI_FINALIZE(IERR)
  STOP
END PROGRAM EXAMPLE21

SUBROUTINE MPIINIT
  USE  MPI
  CALL MPI_INIT( IERR )
  CALL MPI_COMM_RANK(MPI_COMM_WORLD,MYID,IERR)
  CALL MPI_COMM_SIZE(MPI_COMM_WORLD,TOTPS,IERR)
  RETURN
END SUBROUTINE MPIINIT

SUBROUTINE JOBINIT
  USE MPI
  USE JOBDSTRB
  USE WORK
  IF(TOTPS.LT.2) THEN
     WRITE(*,*)'No slave: ', TOTPS
     CALL MPI_FINALIZE(IERR)
     STOP
  ENDIF  
  IF(MYID.EQ.0) NJOB=20
  CALL LOAD
  CALL MPI_BCAST(NJOB,1, MPI_INTEGER,0,MPI_COMM_WORLD,IERR)
  IF(NJOB.LT.1) THEN
     WRITE(*,*)'BAD job number:', NJOB
     CALL MPI_FINALIZE(IERR)
     STOP
  ENDIF  
  ALLOCATE(IDID(NJOB))
  ALLOCATE(NIDID(TOTPS-1))
  ALLOCATE(IGAVE(NJOB))
  ALLOCATE(SIDID(NJOB))
  ALLOCATE(SNIDID(TOTPS-1))
  ALLOCATE(SIFNSH(TOTPS-1))
  ALLOCATE(IFNSH(TOTPS-1))
  ALLOCATE(PFNSH(TOTPS-1))
  ALLOCATE(NIGAVE(TOTPS-1))
  IDID=0
  NIDID=0
  IGAVE=0
  IFNSH =0
  SIDID=0
  SNIDID=0
  SIFNSH =0
  PFNSH =0
  NIGAVE=0
  RETURN
END SUBROUTINE JOBINIT

SUBROUTINE DEMO21
  USE MPI
  USE JOBDSTRB
  INTEGER:: MPITAG, I, J, K, IC
  INTEGER:: MPISTTS(MPI_STATUS_SIZE)
  MPITAG=1


  IF(MYID.NE.0) THEN
     JOB_ASSIGNING: DO
       PRINT*,'Slave #',MYID, ' says: "'   &
                 , 'Master, I am idle. "'
       CALL MPI_SEND(MYID,1,MPI_INTEGER,0, &
            MPITAG+MYID,MPI_COMM_WORLD,IERR)
       CALL MPI_RECV(K,1,MPI_INTEGER,0,    &
                MPITAG+MYID,MPI_COMM_WORLD,&
                MPISTTS,IERR)
       IF(K.LE.NJOB) THEN
          PRINT*,'Slave #',MYID, ' says: ',&
              '"OK, I will be doing job #',&
                        K, ' immediately. "'
          CALL ACALCULATIONJOB(K,MYID)
          IDID(K)=MYID
          NIDID(MYID)= NIDID(MYID)+1
       ELSE
          PRINT*,'Slave #',MYID, ' says: ',&
                 '"Thanks! Master, I am ', &
                 'going home."'
          IFNSH(MYID)=IFNSH(MYID)+1
          EXIT JOB_ASSIGNING
       END IF
     END DO JOB_ASSIGNING


  ELSE
     DO I=1,NJOB+TOTPS-1
        CALL MPI_RECV(IC,1, MPI_INTEGER,   &
            MPI_ANY_SOURCE, MPI_ANY_TAG,   &
            MPI_COMM_WORLD,MPISTTS,IERR)
        CALL MPI_SEND(I,1,MPI_INTEGER,IC,  &
            MPITAG+IC,MPI_COMM_WORLD,IERR)
        IF(I.LE.NJOB) THEN
           PRINT*,'Master     says: "Slave #'& 
                    ,IC,', do job # ',I, '! "'
           IGAVE(I)=IC
           NIGAVE(IC)=NIGAVE(IC)+1
        ELSE
           PRINT*,'Master     says: "Slave #'& 
                 ,IC, ', no more job today! "'
           PFNSH(IC)=PFNSH(IC)+1
        END IF
     END DO
  END IF


  CALL  MPI_REDUCE(IDID,SIDID,NJOB,        &
                   MPI_INTEGER,MPI_SUM,0,  &
                   MPI_COMM_WORLD,IERR)
  CALL  MPI_REDUCE(NIDID,SNIDID,TOTPS-1,   &
                   MPI_INTEGER,MPI_SUM,0,  &
                   MPI_COMM_WORLD,IERR)
  CALL  MPI_REDUCE(IFNSH,SIFNSH,TOTPS-1,   &
                   MPI_INTEGER,MPI_SUM,0,  &
                   MPI_COMM_WORLD,IERR)

  IF(MYID.EQ.0) THEN
     OPEN(1,FILE='msm.out')
     WRITE(1,*)'Total number of jobs:',NJOB
     WRITE(1,*)
     WRITE(1,*)'Total number of processes:',TOTPS,' ,'
     WRITE(1,*)'  of which ',TOTPS-1,' are slaves,' 
     WRITE(1,*)'  and Rank 0 is the Master.'
     WRITE(1,*)
     WRITE(1,*)
     WRITE(1,*)
     WRITE(1,*)
     WRITE(1,*)'Job assigned to and'   &
               , ' completed by slaves '
     WRITE(1,100) 'Job #:  ',(ISLV,ISLV=1, NJOB)
     WRITE(1,100) '#assgn.:',IGAVE
     WRITE(1,100) '#cmplt.:',SIDID
     WRITE(1,*)
     WRITE(1,*)
     WRITE(1,*)'Number of jobs assigned to and' &
                      , ' completed by slaves '
     WRITE(1,100) 'Slave #:',(ISLV,ISLV=1,TOTPS-1)
     WRITE(1,100) '#assgn.:',NIGAVE
     WRITE(1,100) '#cmplt.:',SNIDID
     WRITE(1,*)
     WRITE(1,*)
     WRITE(1,*)'Times sent to and received by' &
                    , ' slaves for job ending'
     WRITE(1,100) 'Slave #:',(ISLV,ISLV=1,TOTPS-1)
     WRITE(1,100) 'Sent:   ',PFNSH
     WRITE(1,100) 'Recved: ',SIFNSH
     100 FORMAT(1X,a8,10000I5)
     CLOSE(1)
  END IF


  DEALLOCATE(IDID,NIDID,NIGAVE)
  DEALLOCATE(IFNSH,PFNSH,IGAVE)
  DEALLOCATE(SIDID,SNIDID,SIFNSH)


  RETURN
END SUBROUTINE DEMO21

subroutine acalculationjob(number,thread)
  ! do senseless busy-work, report back
  use work
  implicit none
  integer :: number,thread,i,st
  st=sleeptime(number) 
  call sleep(st)
  write(*,101)number,thread,st
101 format(' Job number ',i2,' done by process number ',i2,' sleep for ',i3,' seconds.')
  return
end subroutine acalculationjob

