! This is an example code for user defined
! data types.
! HPCVL, Queen's University, 2006


MODULE MPI
  INCLUDE 'mpif.h'
  INTEGER::MYID,TOTPS, IERR
END MODULE MPI

PROGRAM MPIDATATEST4
  USE   MPI
  CALL  MPIINIT
  CALL  TEST4()
  CALL  MPI_FINALIZE(IERR)
  STOP
END PROGRAM MPIDATATEST4

SUBROUTINE MPIINIT
  USE  MPI
  CALL MPI_INIT( IERR )
  CALL MPI_COMM_RANK(MPI_COMM_WORLD,MYID,IERR)
  CALL MPI_COMM_SIZE(MPI_COMM_WORLD,TOTPS,IERR)
  RETURN
END SUBROUTINE MPIINIT

SUBROUTINE TEST4()
  USE MPI
  use fname

  INTEGER, PARAMETER            :: RL_SIZE=5
  INTEGER, PARAMETER            :: ARRAY_SIZE=5
  INTEGER, PARAMETER            :: BLOCK_SIZE=3

  TYPE A_TYPE
    INTEGER                     :: IN
    REAL                        :: RL(RL_SIZE)
    DOUBLE PRECISION            :: DP
  END TYPE A_TYPE

  TYPE (A_TYPE)                 :: ARRAY(ARRAY_SIZE)

  INTEGER                       :: NEW_REAL
  INTEGER                       :: NEW_A_TYPE
  INTEGER                       :: NEW_A_TYPE_SIZED
  INTEGER                       :: LENTH(BLOCK_SIZE)
  INTEGER                       :: TYPES(BLOCK_SIZE)
  INTEGER(KIND=MPI_ADDRESS_KIND):: DIST( BLOCK_SIZE)

  INTEGER(KIND=MPI_ADDRESS_KIND):: INIT
  INTEGER(KIND=MPI_ADDRESS_KIND):: FINL
  INTEGER(KIND=MPI_ADDRESS_KIND):: INTV
  INTEGER                       :: I, J

  ! Informing MPI the true elemental size of the 
  ! REAL :: RL(:) array, in case it is allocated
  ! differently from MPI_REAL by the compiler.
  CALL MPI_GET_ADDRESS(ARRAY(1)%RL(1),INIT,IERR)
  CALL MPI_GET_ADDRESS(ARRAY(1)%RL(2),FINL,IERR)
  INTV=FINL-INIT
  CALL MPI_TYPE_CREATE_RESIZED(MPI_REAL, 0, INTV, &
                               NEW_REAL,    IERR)
  CALL MPI_TYPE_COMMIT(NEW_REAL, IERR)

  ! Constructing an MPI datatype NEW_A_TYPE for 
  ! the ARRAY.
  LENTH(1)=1
  LENTH(2)=RL_SIZE
  LENTH(3)=1
  TYPES(1)=MPI_INTEGER
  TYPES(2)=NEW_REAL
  TYPES(3)=MPI_DOUBLE_PRECISION
  CALL MPI_GET_ADDRESS(ARRAY(1)%IN,    DIST(1), IERR)
  CALL MPI_GET_ADDRESS(ARRAY(1)%RL(1), DIST(2), IERR)
  CALL MPI_GET_ADDRESS(ARRAY(1)%DP,    DIST(3), IERR)
  DIST(3)=DIST(3)-DIST(1)
  DIST(2)=DIST(2)-DIST(1)
  DIST(1)=0

  CALL MPI_TYPE_CREATE_STRUCT(BLOCK_SIZE,LENTH,DIST,&
                              TYPES,NEW_A_TYPE,IERR)
  CALL MPI_TYPE_COMMIT(NEW_A_TYPE, IERR)

  ! Informing MPI the true elemental size of the 
  ! MPI datatype NEW_A_TYPE, in case it is made 
  ! different by the compiler from MPI thought.
  CALL MPI_GET_ADDRESS(ARRAY(1)%IN, INIT, IERR)
  CALL MPI_GET_ADDRESS(ARRAY(2)%IN, FINL, IERR)
  INTV=FINL-INIT

  CALL MPI_TYPE_CREATE_RESIZED(NEW_A_TYPE, 0, &
                 INTV, NEW_A_TYPE_SIZED, IERR)
  CALL MPI_TYPE_COMMIT(NEW_A_TYPE_SIZED, IERR)

  OPEN(10,FILE=FILENAME(MYID))

  ! For testing, assign different values to 
  ! ARRAYs among different processes, then
  ! broadcast, and check them.
  WRITE(10,*)
  WRITE(10,*)' Rank ',MYID,':'
  WRITE(10,*)' Before Broadcast ------------------------'
  CALL MPI_BARRIER(MPI_COMM_WORLD,IERR)
  DO I=1, ARRAY_SIZE
     IF(MYID.EQ.0) THEN
        ARRAY(I)%IN=I
        ARRAY(I)%DP=I*10.0D0
        DO J=1, RL_SIZE
           ARRAY(I)%RL(J)=J+I*100+0.8D0
        END DO
     ELSE
        ARRAY(I)%IN=-1
        ARRAY(I)%RL=-1.0
        ARRAY(I)%DP=-1.0D0
     END IF
     WRITE(10,101) ARRAY(I)%IN,ARRAY(I)%RL,ARRAY(I)%DP,MYID
  END DO

  CALL MPI_BCAST(ARRAY(1)%IN,ARRAY_SIZE,        &
           NEW_A_TYPE_SIZED,0,MPI_COMM_WORLD,IERR)
  CALL MPI_BARRIER(MPI_COMM_WORLD,IERR)

  WRITE(10,*)' After Broadcast ------------------------'
  DO I=1, ARRAY_SIZE
     WRITE(10,101)ARRAY(I)%IN,ARRAY(I)%RL,ARRAY(I)%DP,MYID
  END DO

  CLOSE(10)
101 FORMAT(1X,I3,6F10.2," on rank",I2)
  CALL MPI_TYPE_FREE(NEW_REAL,         IERR)
  CALL MPI_TYPE_FREE(NEW_A_TYPE,       IERR)
  CALL MPI_TYPE_FREE(NEW_A_TYPE_SIZED, IERR)

  RETURN
END SUBROUTINE TEST4

