/* This a "stripped" version of darray.c by Zdzislaw Meglicki
 * from http://beige.ucs.indiana.edu/I590/node102.html
 * Hartmut Schmider December 2013 */

#include <stdio.h>     
#include <stdlib.h>    
#include <unistd.h>    
#include <string.h>    
#include <sys/types.h> 
#include <sys/stat.h>  
#include <mpi.h>       

#define TRUE 1
#define FALSE 0
#define BOOLEAN int
#define SIZE 512

int report(int r,int s,int *g, int *d, int *a, int *p, int o);

int main(int argc, char **argv){
    /* my variables */
    int my_rank, pool_size, i, order, file_name_length,
	array_of_gsizes[3], array_of_distribs[3], array_of_dargs[3], array_of_psizes[3],
	*write_buffer, write_buffer_size, count, *read_buffer, read_buffer_size;
    BOOLEAN my_read_error = FALSE, read_error = FALSE;
    char *file_name = NULL, message[BUFSIZ];
    
    /* MPI variables */
    MPI_Offset file_size;
    MPI_File fh;
    MPI_Status status;
    MPI_Datatype file_type;
    MPI_Aint file_type_extent;
    int file_type_size;
    
    /* getopt variables */
    extern char *optarg;
    int c;
    
    /* error handling variables */
    extern int errno;
    
    /* ACTION */
    MPI_Init(&argc,&argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &pool_size);

    if (my_rank==0) {
	file_name="darray.dat";
	file_name_length = strlen(file_name) + 1;
	printf("file_name         = %s\n", file_name);
	printf("file_name_length  = %d\n", file_name_length);
    } 
    
    MPI_Bcast(&file_name_length, 1, MPI_INT, 0, MPI_COMM_WORLD);

    if (my_rank!=0) file_name = (char*) malloc(file_name_length);

    MPI_Bcast(file_name, file_name_length, MPI_CHAR, 0, MPI_COMM_WORLD);
    
    /* Prepare for calling MPI_Type_create_darray */
    for (i = 0; i<3; i++) {
	array_of_gsizes[i] = SIZE;
	array_of_distribs[i] = MPI_DISTRIBUTE_BLOCK;
	array_of_dargs[i] = MPI_DISTRIBUTE_DFLT_DARG;
	array_of_psizes[i] = 0;
    }
    MPI_Dims_create(pool_size, 3, array_of_psizes);
    order = MPI_ORDER_C;

    for (i=0; i<pool_size; i++){ /* write info in rank order */
	if (my_rank==i)
	    report(my_rank,pool_size,array_of_gsizes,array_of_distribs,array_of_dargs,array_of_psizes,order);
	MPI_Barrier(MPI_COMM_WORLD);
    }
    /* Now call MPI_Type_create_darray */
    MPI_Type_create_darray(pool_size, my_rank, 3,
			   array_of_gsizes, array_of_distribs,
			   array_of_dargs, array_of_psizes, order,
			   MPI_INT, &file_type);
    MPI_Type_commit(&file_type);
    
    /* Explore the returned type */
    MPI_Type_extent(file_type, &file_type_extent);
    MPI_Type_size(file_type, &file_type_size);
    printf("%3d: file_type_size   = %d\n", my_rank, file_type_size);
    printf("%3d: file_type_extent = %ld\n", my_rank, file_type_extent);
    
    /* Allocate space for your own write buffer based on the
       return of the MPI_Type_size call. */
    write_buffer_size = file_type_size / sizeof(int);
    write_buffer = (int*) malloc(write_buffer_size * sizeof(int));
    
    /* We wait here in case some procs have problems with malloc. */
    MPI_Barrier(MPI_COMM_WORLD);  
    
    /* Initialize the buffer */
    for (i = 0; i < write_buffer_size; i++) 
	*(write_buffer + i) = my_rank * SIZE + i;
    MPI_File_open(MPI_COMM_WORLD, file_name, 
		  MPI_MODE_CREATE | MPI_MODE_WRONLY,
		  MPI_INFO_NULL, &fh);
	
    if (my_rank==0)
	chmod(file_name, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

    MPI_Barrier(MPI_COMM_WORLD); /* sync Barrier */
    
    MPI_File_set_view(fh, 0, MPI_INT, file_type, "native", MPI_INFO_NULL);
    MPI_File_write_all(fh, write_buffer, write_buffer_size, MPI_INT, &status);
    MPI_Get_count(&status, MPI_INT, &count);
    MPI_File_get_size(fh, &file_size);
    printf("%3d: wrote %d integers\n", my_rank, count);
    printf("%3d: file size is %lld bytes\n", my_rank, file_size);
    MPI_File_close(&fh);
    
    /* We have managed to open, write on it and close the file.
       Now we're going to read it the same way we wrote it. */
    read_buffer_size = write_buffer_size;
    read_buffer = (int*) malloc(read_buffer_size * sizeof(int));
    
    MPI_Barrier(MPI_COMM_WORLD); /* Sync Barrier */
    
    MPI_File_open(MPI_COMM_WORLD, file_name, MPI_MODE_DELETE_ON_CLOSE|MPI_MODE_RDONLY,MPI_INFO_NULL, &fh);
    MPI_File_set_view(fh, 0, MPI_INT, file_type, "native", MPI_INFO_NULL);
    MPI_File_read_all(fh, read_buffer, read_buffer_size, MPI_INT, &status);
    MPI_Get_count(&status, MPI_INT, &count);
    printf("%3d: read %d integers\n", my_rank, count);
    MPI_File_close(&fh);
    
    /* Now check that the integers read are the same as the ones we wrote. */
    for (i = 0; i < read_buffer_size; i++) {
	if (*(write_buffer + i) != *(read_buffer + i)) {
	    printf("%3d: data read different from data written, i = %d\n", my_rank, i);
	    my_read_error = TRUE;
	}
    }
    
    MPI_Reduce (&my_read_error, &read_error, 1, MPI_INT, MPI_LOR, 0, MPI_COMM_WORLD);           
    
    if (my_rank==0) if (! read_error) printf("--> All data read back is correct.\n");
    MPI_Finalize();
    exit(0);
}

int report(int r,int s,int *g, int *d, int *a, int *p, int o){
    printf ("%3d: calling MPI_Type_create_darray with\n", r);
    printf ("%3d:    pool_size         = %d\n", r, s);
    printf ("%3d:    my_rank           = %d\n", r, r);
    printf ("%3d:    ndims             = 3\n", r);
    printf ("%3d:    array_of_gsizes   = (%d, %d, %d)\n", r,g[0],g[1],g[2]);
    printf ("%3d:    array_of_distribs = (%d, %d, %d)\n", r,d[0],d[1],d[2]);
    printf ("%3d:    array_of_dargs    = (%d, %d, %d)\n", r,a[0],a[1],a[2]);
    printf ("%3d:    array_of_psizes   = (%d, %d, %d)\n", r,p[0],p[1],p[2]);
    printf ("%3d:    order             = %d\n", r, o);
    printf ("%3d:    type              = %d\n", r, MPI_INT);
    return 0;
}

