program collective

  use mpi
  use mpi_basics

  ! Demonstrating Collective IO based on Using MPI-2 p.65, Fig.3.7 */

  implicit none

  integer :: etype=MPI_INTEGER,ftype,file,mode=ior(MPI_MODE_DELETE_ON_CLOSE,ior(MPI_MODE_CREATE,MPI_MODE_WRONLY))
  integer (kind=MPI_OFFSET_KIND) :: offset
  integer :: i,nints,nbuf,FILESIZE=1048576,INTS_PER_BLK=16
  integer, allocatable :: buf(:)

  double precision :: t1, t2;

  call mpistart

  nbuf=FILESIZE/size
  nints=nbuf/4
  allocate(buf(nints))

  do i=1,nints
     buf(i)=nints*rank+i
  end do

  ! writing 4 times (mem/disk and ind/col)
  ! creating a 'vector' non-contiguous datatype and committing it
  call MPI_Type_vector(nints/INTS_PER_BLK, INTS_PER_BLK, INTS_PER_BLK*size, etype, ftype, ierr)
  call MPI_Type_commit(ftype, ierr)
  offset=INTS_PER_BLK*4*rank

  ! 1. Independently to memory
  call MPI_File_open(MPI_COMM_WORLD, "/tmp/dump.dat", mode, MPI_INFO_NULL, file, ierr)
  call MPI_File_set_view(file, offset, etype, ftype, "native", MPI_INFO_NULL, ierr)
  call cpu_time(t1)
  call MPI_File_write(file, buf, nints, MPI_INTEGER, MPI_STATUS_IGNORE, ierr)
  call cpu_time(t2)
  call MPI_File_close(file, ierr)
  write(*,'("Time for dumping data = ",f12.8," on rank ",i2," (individual, memory)")')t2-t1, rank
  call MPI_Barrier(MPI_COMM_WORLD, ierr)

  ! 2. Collectively to memory
  call MPI_File_open(MPI_COMM_WORLD, "/tmp/dump.dat", mode, MPI_INFO_NULL, file, ierr)
  call MPI_File_set_view(file, offset, etype, ftype, "native", MPI_INFO_NULL, ierr)
  call cpu_time(t1)
  call MPI_File_write_all(file, buf, nints, MPI_INTEGER, MPI_STATUS_IGNORE, ierr)
  call cpu_time(t2)
  call MPI_File_close(file, ierr)
  write(*,'("Time for dumping data = ",f12.8," on rank ",i2," (collective, memory)")')t2-t1, rank
  call MPI_Barrier(MPI_COMM_WORLD, ierr)

  ! 3. Independently to disk
  call MPI_File_open(MPI_COMM_WORLD, "dump.dat", mode, MPI_INFO_NULL, file, ierr)
  call MPI_File_set_view(file, offset, etype, ftype, "native", MPI_INFO_NULL, ierr)
  call cpu_time(t1)
  call MPI_File_write(file, buf, nints, MPI_INTEGER, MPI_STATUS_IGNORE, ierr)
  call cpu_time(t2)
  call MPI_File_close(file, ierr)
  write(*,'("Time for dumping data = ",f12.8," on rank ",i2," (individual, disk)")')t2-t1, rank
  call MPI_Barrier(MPI_COMM_WORLD, ierr)

  ! 4. Collectively to memory
  call MPI_File_open(MPI_COMM_WORLD, "dump.dat", mode, MPI_INFO_NULL, file, ierr)
  call MPI_File_set_view(file, offset, etype, ftype, "native", MPI_INFO_NULL, ierr)
  call cpu_time(t1)
  call MPI_File_write_all(file, buf, nints, MPI_INTEGER, MPI_STATUS_IGNORE, ierr)
  call cpu_time(t2)
  call MPI_File_close(file, ierr)
  write(*,'("Time for dumping data = ",f12.8," on rank ",i2," (collective, disk)")')t2-t1, rank
  call MPI_Barrier(MPI_COMM_WORLD, ierr)

  ! wrap it up
  deallocate(buf)
  call MPI_Finalize(ierr)
  stop

end program collective

