! This a "stripped" version of darray.c by Zdzislaw Meglicki
! from http://beige.ucs.indiana.edu/I590/node102.html
! Hartmut Schmider January 2014 

program darray

  use mpi

  integer, parameter :: SIZE=512

  integer :: my_rank, pool_size, i, order, file_name_length
  integer, dimension (3) :: array_of_gsizes, array_of_distribs, array_of_dargs, array_of_psizes
  integer, dimension(:), allocatable :: write_buffer, read_buffer 
  integer :: write_buffer_size, read_buffer_size,count
  logical :: my_read_error=.FALSE., read_error=.FALSE.
  character*80 :: file_name
  
  integer (kind=MPI_OFFSET_KIND) :: file_size
  integer :: fh, file_type, file_type_size, file_type_extent, errno, ierr
  integer, dimension(MPI_STATUS_SIZE) :: status
  integer (kind=MPI_OFFSET_KIND) :: disp=0

  ! ACTION
  call MPI_Init(ierr)
  call MPI_Comm_rank(MPI_COMM_WORLD, my_rank, ierr)
  call MPI_Comm_size(MPI_COMM_WORLD, pool_size, ierr)

  if (my_rank.eq.0) then
     file_name="darray.dat"
     file_name_length=10
     write(*,*)"file_name         = ", file_name
     write(*,*)"file_name_length  = ", file_name_length
  end if

  call MPI_Bcast(file_name_length,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
  call MPI_Bcast(file_name, file_name_length, MPI_CHARACTER, 0, MPI_COMM_WORLD, ierr)
    
  ! Prepare for calling MPI_Type_create_darray 
  do i=1,3
     array_of_gsizes(i) = SIZE;
     array_of_distribs(i) = MPI_DISTRIBUTE_BLOCK;
     array_of_dargs(i) = MPI_DISTRIBUTE_DFLT_DARG;
     array_of_psizes(i) = 0;
  end do

  call MPI_Dims_create(pool_size, 3, array_of_psizes, ierr)
  order = MPI_ORDER_F;

  do i=0,pool_size-1  ! write info in rank order 
     if (my_rank.eq.i) call report(my_rank,pool_size, &
          array_of_gsizes,array_of_distribs,array_of_dargs,array_of_psizes,order);
     call MPI_Barrier(MPI_COMM_WORLD,ierr);
  end do

  ! Now call MPI_Type_create_darray
  call MPI_Type_create_darray(pool_size, my_rank, 3, &
       array_of_gsizes, array_of_distribs, &
       array_of_dargs, array_of_psizes, order, &
       MPI_INTEGER, file_type,ierr)
  call MPI_Type_commit(file_type,ierr)
  
  ! Explore the returned type 
  call MPI_Type_extent(file_type, file_type_extent,ierr)
  call MPI_Type_size(file_type, file_type_size,ierr)
  write(*,*)my_rank,": file_type_size   = ", file_type_size
  write(*,*)my_rank,": file_type_extent = ", file_type_extent

  ! Allocate space for your own write buffer based on the
  ! return of the MPI_Type_size call.
  write_buffer_size = file_type_size / sizeof(int);
  allocate (write_buffer(write_buffer_size))
   
  ! We wait here in case some procs have problems with allocation
  call MPI_Barrier(MPI_COMM_WORLD,ierr)
    
  ! Initialize the buffer
  do i=1,write_buffer_size
     write_buffer(i) = my_rank * SIZE + i
  end do

  call MPI_File_open(MPI_COMM_WORLD, file_name, ior(MPI_MODE_CREATE,MPI_MODE_WRONLY), &
       MPI_INFO_NULL, fh, ierr)
	
  ! if (my_rank==0)
  ! chmod(file_name, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

  call MPI_Barrier(MPI_COMM_WORLD,ierr)
    
  call MPI_File_set_view(fh, disp, MPI_INTEGER, file_type, "native", MPI_INFO_NULL, ierr)
  call MPI_File_write_all(fh, write_buffer, write_buffer_size, MPI_INTEGER, status, ierr)
  call MPI_Get_count(status, MPI_INTEGER, count, ierr)
  call MPI_File_get_size(fh, file_size, ierr)
  write(*,*)my_rank,": wrote ",count," integers"
  write(*,*)my_rank,": file size is ",file_size," bytes"
  call MPI_File_close(fh, ierr)
    
   ! We have managed to open, write on it and close the file.
   ! Now we're going to read it the same way we wrote it.
    read_buffer_size = write_buffer_size
    allocate (read_buffer(read_buffer_size))
    
    call MPI_Barrier(MPI_COMM_WORLD,ierr) ! Sync Barrier
    
    call MPI_File_open(MPI_COMM_WORLD, file_name, ior(MPI_MODE_DELETE_ON_CLOSE,MPI_MODE_RDONLY),MPI_INFO_NULL, fh, ierr)
    call MPI_File_set_view(fh, disp, MPI_INTEGER, file_type, "native", MPI_INFO_NULL, ierr)
    call MPI_File_read_all(fh, read_buffer, read_buffer_size, MPI_INTEGER, status, ierr)
    call MPI_Get_count(status, MPI_INTEGER, count, ierr)
    write(*,*)may_rank,": read ",count," integers"
    call MPI_File_close(fh, ierr)
    
    ! Now check that the integers read are the same as the ones we wrote.
    do i=1,read_buffer_size
       if (write_buffer(i).ne.read_buffer(i)) then
          write(*,*)my_rank,": data read different from data written, i = ",i
          my_read_error = .TRUE.
       end if
    end do

    call MPI_Reduce (my_read_error,read_error,1,MPI_LOGICAL,MPI_LOR,0,MPI_COMM_WORLD,ierr)     
    
    if (my_rank.eq.0) then
       if (.not.read_error) write(*,*)"--> All data read back is correct."
    end if

    call MPI_Finalize(ierr)
    stop
  end program darray


subroutine report(r,s,g,d,a,p,o)
  integer, intent(in) :: r,s,o
  integer, dimension(3), intent(in) :: g,d,a,p

  write(*,*)r,": calling MPI_Type_create_darray with"
  write(*,*)r,": pool_size         = ",s
  write(*,*)r,": my_rank           = ",r
  write(*,*)r,": ndims             = 3"
  write(*,*)r,": array_of_gsizes   = (",g(1),",",g(2),",",g(3),")"
  write(*,*)r,": array_of_distribs = (",d(1),",",d(2),",",d(3),")"
  write(*,*)r,": array_of_dargs    = (",a(1),",",a(2),",",a(3),")"
  write(*,*)r,": array_of_psizes   = (",p(1),",",p(2),",",p(3),")"
  write(*,*)r,": order             = ",o
  write(*,*)r,": type              = ",MPI_INTEGER
end subroutine report


