program outin
  
  use mpi
  use mpi_basics
  
  implicit none
  
  integer :: contig,etype=MPI_INTEGER,ftype,file
  integer (kind=MPI_ADDRESS_KIND) :: offset, extent=32
  integer :: i,irank,nbuf=20
  integer, allocatable :: buf(:)
  
  call mpistart
  allocate (buf(nbuf))
  do i=1,nbuf
     buf(i)=nbuf*rank+i
  end do

  ! writing
  call MPI_File_open(MPI_COMM_WORLD, "outin.dat", ior(MPI_MODE_CREATE,MPI_MODE_WRONLY), MPI_INFO_NULL, file, ierr)
  ! creating a datatype with two integers and a 'hole' of 6 integers 
  call MPI_Type_contiguous(2,etype,contig,ierr)
  offset=0
  call MPI_Type_create_resized(contig,offset,extent,ftype,ierr)
  call MPI_Type_commit(ftype,ierr)
  ! the actual write is non-contiguous using the new datatype 
  offset=rank*8
  call MPI_File_set_view(file,offset,etype,ftype,"native",MPI_INFO_NULL,ierr)
  call MPI_File_write(file, buf, nbuf, MPI_INT, MPI_STATUS_IGNORE,ierr)
  ! closing the file and syncing after write is done 
  call MPI_File_close(file,ierr)
  call MPI_Barrier(MPI_COMM_WORLD,ierr)
    
  ! reading 
  call MPI_File_open(MPI_COMM_WORLD, "outin.dat", MPI_MODE_RDONLY, MPI_INFO_NULL, file, ierr)
  ! the read is contiguous, 'scrambling' the data 
  offset=nbuf*4*rank
  call MPI_File_set_view(file,offset,MPI_INT,MPI_INT,"native",MPI_INFO_NULL, ierr)
  call MPI_File_read(file, buf, nbuf, MPI_INT, MPI_STATUS_IGNORE, ierr)
  call MPI_File_close(file,ierr)

  ! output to see the result (force order) 
  do irank=0,size-1
     call MPI_Barrier(MPI_COMM_WORLD,ierr)
     if (irank.eq.rank) then 
        do i=1,nbuf 
           write(*,*)" Process #",rank," Element ",i," = ",buf(i)
        end do
     end if
     call MPI_Barrier(MPI_COMM_WORLD,ierr)
  end do
  
  ! wrap it up 
  call MPI_Type_free(ftype,ierr);
  call MPI_Finalize(ierr);
  
end program outin
