program pseudoO
  ! Pseudo-Parallel output, each process to its own file
  ! Hartmut Schmider December 2013

  use mpi
  use mpi_basics

  implicit none
  double precision, allocatable :: stuff(:)
  double precision :: t1,t2
  integer :: ndata,nmax,i,j,ransize,handle
  integer, dimension(MPI_STATUS_SIZE) :: status
  integer (kind=MPI_OFFSET_KIND) :: offset
  integer, allocatable :: seed(:)
  
  call mpistart

  if(rank.eq.0)then
     write(*,*)' How big should the data be ?'
     read(*,*)ndata
     call cpu_time(t1)
  end if

! Broadcasting size, loading elements with random numbers
  call MPI_bcast(ndata,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
  nmax=ndata/size
  allocate(stuff(nmax))
  call random(stuff,nmax,rank)
  if (rank.eq.0) then
     call cpu_time(t2)
     write(*,*)' Time for loading array = ',t2-t1,' seconds'
  end if
  
! Dumping data individually per process (in a single file)
  if (rank.eq.0) write(11,*)nmax
  offset=rank*8*nmax
  call MPI_File_open(MPI_COMM_WORLD,"/tmp/fort.10",ior(MPI_MODE_CREATE,MPI_MODE_WRONLY),MPI_INFO_NULL,handle,ierr)
  call MPI_File_seek(handle,offset,MPI_SEEK_SET,ierr)
  call MPI_File_write(handle,stuff,nmax,MPI_DOUBLE_PRECISION,status,ierr)
  call MPI_File_close(handle,ierr)
  if (rank.eq.0) then
     call cpu_time(t1)
     write(*,*)' Time for dumping data = ',t1-t2,' seconds'
  end if
! Wrapping up  
  deallocate(stuff)
  call MPI_finalize(ierr)
  stop
end program pseudoO
